//*****************************************************************************
//      C++ functions for:
//		Drawing of 1D and 3D representations of hydrogen orbitals
//
//  		Original h3d and XXXwf subroutines written in FORTRAN by
//      	Jeff Reimers and Georger Bacskay,
//		(School of Chemistry, University of Sydney)
//
//      Written/Adapted by Stuart Prescott:
//              Written/Adapted:        12/09/1996
//              Last Modified:          27/10/1996
//
//      Libraries Required:
//              libsx (Dominic Giamapolo - Athena Widget Set)
//
//      System headers:
//              Streamed I/O
//              Formatted I/O
//              Maths fns
//
//      Other Object Classes used:
//              template arrays
//              3D points / vectors / matrices
//              graphs
//              colours
//              fonts
//              libsx-extensions
//              maths extensions
//              main program data structure
//              orbital id and name
//		wavefunction control addon
//
//*****************************************************************************

#include <math.h>		//functions req'd for this file
#include <stdio.h>
#include <iostream.h>

extern "C"
{
    #include <libsx.h>
}

#include "h3d.h"
#include "arraytem.h"
#include "vectors.h"
#include "mathxtra.h"	
#include "graph.h"
#include "main.h"		//declares the types . . .
#include "psi2.h"
#include "grafxten.h"

#include "coldefs.h"
#include "fontdefs.h"

// 1 to turn on debug writes; 0 to turn off . . .
#define DEBUG1D 	0
#define DEBUG		0
#define DEBUGINT	0
#define DEBUGWF 	0
#define DEBUGSTR	0
#define DEBUGRESCALE	0
#define DEBUGPAL	0
#define DEBUGPALUSE	0
#define DEBUGDLEV	0
#define DEBUGSHOW	0
#define DEBUGAXESCHECK	0
#define DEBUGDYNALEVEL	0

//normally 1 . . .
#define GETFONTINFO	1
#define AXESCHECKSAFETY 1

extern FontDefs *fonts;
extern ColourDefs *pal;

//****************************************************************************
//****************************************************************************
void SetupOrbitals(MyProgram *me)
{
    // *** prepare all necessary values for the orbitals

    SetDrawArea(me->mainwin.draw_area);

    me->mainwin.orbital.gcol=GetPrivateColor();
    if (me->mainwin.orbital.gcol==-1)
    {
	ColourError();
	return;
    }
    SetPrivateColor(me->mainwin.orbital.gcol, 0x00, 0xE0, 0xA3);

    //inital modes == selections and density . . .
    me->mode = WAITING_FOR_SELECTION;
    me->mainwin.orbital.fntype=2;

    return;
}

//****************************************************************************
//****************************************************************************
void OrbitalView::Show (OrbitalDescriptor *o, int i, matrix *trans,
                	int graphcol, Widget w, Widget wt, int x, int y) 
{
    // Set up the environment for the orbital - 
    // Draw only those bits which need to be draw.

    gcol=graphcol;
    draw_area=w;
    title_area=wt;
    
    int refreshmode=0;		//only true if there are changed conditions
				//eg viewpoint, wfval, which orbital . . .

    #if DEBUGSHOW
        if ((xpix!=x)||(ypix!=y))
        {
	    cout << "screen size changed "<< x<<" "<<y<<"\n";
        } 
    #endif

    xpix=x;
    ypix=y;

    #if DEBUGPAL     // do a palette check
    {
	for (int color=0; color<= (pal->ngcol); ++color)
	{
	    SetColor(pal->rpal->get(color));
	    DrawText("[", 620, 4*(color+1));
	}
	for (int color=0; color<= (pal->nrcol); ++color)
	{
	    SetColor(pal->gpal->get(color));
	    DrawText("[", 630, 4*(color+1));
	}
	SetColor (gcol);
	DrawText("XX", 620, 300);
	SetColor (pal->white);
	DrawText("W", 630, 320);
	SetColor(pal->black);
	DrawText("B", 630, 325);

	cout <<"Colour Test\n";
    }
    #endif

    //Split up the tasks according to what has changed.

    int CallTitleAll    = FALSE;
    int CallSetup1d	= FALSE;
    int CallSetup3d     = FALSE;
    int CallClear3d	= FALSE;
    int CallClearAll	= FALSE;
    int CallDraw1d	= FALSE;
    int CallDraw3d	= FALSE;
    int CallRedraw3d    = FALSE;
    int CallDrawLev	= FALSE;

    if ( i!=itype)
    {
	itype = i;
    	CallTitleAll    = TRUE;
    	CallSetup1d     = TRUE;
    	CallSetup3d     = TRUE;
    	CallClearAll    = TRUE;
    	CallDraw1d      = TRUE;
    	CallDraw3d      = TRUE;
    	CallDrawLev     = TRUE;
    }

    if ( (o->n!=id.n) || (o->l!=id.l) || (o->m!=id.m) )
    {
	id = *o;
	char *s1=lwf;
	char *s2=id.str;
	while ((*s1++=*s2++));

        #if DEBUGSHOW
            cout << "nlm changed\n";
            cout << "n,l,m= "<<id.n<<" "<<id.l<<" "<<id.m
                 << "\n" << *(trans);
        #endif

	CallClearAll = TRUE;
	CallTitleAll = TRUE;
	CallSetup1d  = TRUE;
	CallSetup3d  = TRUE;
	CallDraw1d   = TRUE;
	CallDraw3d   = TRUE;
	CallDrawLev  = TRUE;
	refreshmode  = 1;
    }

    if (t != *(trans))
    {
	#if DEBUGSHOW
	    cout << "T changed\n";
	    cout << t <<"\n" << *(trans) <<"\n";
	#endif

	t = *(trans);

	CallClear3d = TRUE;
	CallSetup3d = TRUE;
	CallDraw3d  = TRUE;

        refreshmode = 1;
    }

    if (wfmax0!=lastwfval)
    {
	#if DEBUGSHOW
	    cout << "surface level changed\n";
	#endif
	wfmax0 = lastwfval;
        CallClear3d = TRUE;
        CallDraw3d  = TRUE;

	refreshmode = 1;
    }

    if (!refreshmode)
    {
        #if DEBUGSHOW
            cout << "nlm unchanged; T unchanged; REDISPLAY\n";
        #endif

    	CallClearAll = TRUE;
    	CallTitleAll = TRUE;
    	CallDraw1d   = TRUE;
	CallRedraw3d = TRUE;
        CallDrawLev  = TRUE;
    }

    if (CallClearAll) ClearAll();
    if (CallTitleAll) TitleAll();
    if (CallSetup1d)  Setup1d();
    if (CallSetup3d)  Setup3d();
    if (CallClear3d)  Clear3d();
    if (CallDraw1d)   Draw1d();
    if (CallDraw3d)   Draw3d();
    if (CallRedraw3d) Redraw3d();
    if (CallDrawLev)  DrawLev();


    return;
}

//****************************************************************************
//****************************************************************************
void OrbitalView::ClearAll()
{
    ClearDrawArea();
    SetBgColor(draw_area, pal->black);
    SetColor(pal->black);
    DrawFilledBox(0,0, xpix, ypix);
    return;
}

//****************************************************************************
//****************************************************************************
void OrbitalView::TitleAll()
{
    // *** put the title in the title_area (using the redisplay function)

    SetDrawArea(title_area);
    TitleRedisplay (title_area, xpix, ypix, this);

    SetDrawArea(draw_area);
    SetWidgetFont(draw_area, fonts->menu);
    return;
}

//****************************************************************************
//****************************************************************************
void OrbitalView::Setup1d()
{
    // prepare for the 1D graphs  . . .
    // - do all calculations and keep them handy for redraw speed.

    #if DEBUGSHOW
	cout << "Setup1D\n";
    #endif
	
    double delr, rmax, rwfmax,twfmax,pwfmax;

    int i, j;

    #if GETFONTINFO
	font=fonts->plottitle;
	fonth= FontHeight(font);
    #else
	fonth=10;
    #endif

    //************************** data for 1D plots ***************************

    //**** radial function ****
    rmax= pow(2.0, (id.n+1));
    delr= rmax / mrr;
    rwfmax= 0.0;
    j= 0;
    for (i= mrr; i>=0; --i)
    {
        r1->set(i, i * delr);
        wfr->set(i, radwf(id.n,id.l,r1->get(i),1.0));
        if (itype == 2) wfr->set(i, pow(wfr->get(i),2));
	//**** get value of radial wf at outermost turning point ****
        if ( (rwfmax==0.0) && (i != mrr) &&
                ( abs(wfr->get(i+1)) > abs(wfr->get(i)) ) )
        {
            rwfmax= wfr->get(i+1);
            rexpand= r1->get(i+1);
        }
    }
    if (id.n == 1)
    {
        i= mrr/4;
        rexpand= r1->get(i);
        rwfmax= wfr->get(i);
    }

    // **** theta ****
    delr= pi / mrr;
    twfmax= 0.0;
    for (i= 0; i<= mrr; ++i)
    {
        theta1->set(i, i * delr);
        wft->set(i, angwf(id.l, abs(id.m), 
		    cos(theta1->get(i)), sin(theta1->get(i))) );
        if (itype == 2) wft->set(i, pow(wft->get(i),2) );
        //**** get max value of theta ****
        if (abs(wft->get(i)) > abs(twfmax))
        {
          twfmax= wft->get(i);
          texpand= theta1->get(i);
        }
    }
    if (texpand == 0.0) texpand= theta1->get(1);

    //**** phi ****
    pwfmax= 0.0;
    for (i= 0; i<=mrr; ++i)
    {
        phi1->set(i, i * delr * 2.0);
        if (id.m == 0)
        {
            wfp->set(i, 1/(rt2*rtpi));
        }
        else
        {
            if (id.m > 0)
            {
                wfp->set(i, sin(phi1->get(i)*id.m) / rtpi);
            }
            else
            {
                wfp->set(i, cos(phi1->get(i)*id.m) / rtpi);
            }
        }
        if (itype == 2) wfp->set(i, pow(wfp->get(i),2) );

        //**** get max value of phi ****
        if (abs(wfp->get(i)) > abs(pwfmax))
        {
            pwfmax= wfp->get(i);
            pexpand= phi1->get(i);
        }
    }

    //**** starting surface level, vert scale max in plot ****
    wfmax= rwfmax * twfmax * pwfmax;
    #if DEBUG1D
        cout<<"WFmax= "<<wfmax<<"="<<rwfmax<<" * "<<twfmax<<" * "<<pwfmax<<"\n";
    #endif
    if (wfmax < 0.001)
    {
        wfscale= 1000.0;
    }
    else
    {
        wfscale= 1.0;
    }
    wfmax0= abs(wfmax) * 0.45;
    wfmax= wfmax0 * wfscale;
    #if DEBUG1D
        cout << "surface level= " << wfmax << " scale= "<< wfscale<<"\n";
    #endif

    //**** rescale all 1D plots for max in other 2 vars ****
    double a= 180.0 / pi;
    double wfrescaletp = twfmax * pwfmax * wfscale;
    double wfrescalepr = pwfmax * rwfmax * wfscale;
    double wfrescalert = rwfmax * twfmax * wfscale;
    for (i= 0; i<=mrr; ++i)
    {
        #if DEBUGRESCALE
	    printf ("%d: %f %f %f %f %f \n", i, theta1->get(i),
					phi1->get(i),
                                        wfr->get(i),
                                        wft->get(i),
                                        wfp->get(i));
	#endif

	theta1->set(i, theta1->get(i) * a);
        phi1->set(i, phi1->get(i) * a);
        wfr->set(i, wfr->get(i) * wfrescaletp);
        wft->set(i, wft->get(i) * wfrescalepr);
        wfp->set(i, wfp->get(i) * wfrescalert);

        #if DEBUGRESCALE
            printf ("   %f %f %f %f %f \n", theta1->get(i),
                                        phi1->get(i),
                                        wfr->get(i),
                                        wft->get(i),
                                        wfp->get(i));
        #endif
    }

    pexpand*=a;
    texpand*=a;

    #if DEBUGRESCALE
	cout << "Rescaling done!\n";
    #endif

    return;
}

//****************************************************************************
//****************************************************************************
void OrbitalView::Draw1d()
{
    // Draw the 1D plots up on the screen
    // using the data calculated in Setup1d()
    #if DEBUGSHOW
	cout << "Draw1D\n";
    #endif
	
    double rmax, vmax,vmin;

    int ca1;


    //*********************** draw 1D plots ********************
    rmax= pow(2.0, (id.n+1));

    font=fonts->plotscale;
    SetWidgetFont(draw_area, font);
    fonth=FontHeight(font);

    vmax= 0.6;
    if (id.n != 1) vmax= axmax(wfmax*4.0);

    #if DEBUG1D
	cout << "vmax= " <<vmax << "\n";
    #endif

    vmin= 0.0;
    ca1= 14;
    if (itype == 1)
    {
        vmin= - vmax;
        ca1= 11;
    }

    //**** radial ****

    int len;
    len = sprintf(str, "AT %s=%g, %s=%g","THETA",(texpand),"PHI",(pexpand));

    g[0].setdrawarea 	(draw_area);
    g[0].setcorner 	(0, int((ypix-5)/2) );
    g[0].setscreensize 	(xpix, ypix);
    g[0].setdomain 	(0.0,rmax);
    g[0].setrange 	(vmin,vmax);
    g[0].setaxislen 	( int (xpix/2 - 6*(fonth+1) -7),
				  int (ypix/2 -12 -4*fonth)      );

    g[0].setaxistitles 	("R (Bohr)", title(itype));
    g[0].setplottitle 	(str);

    g[0].setticlen 	(4, 4);
    g[0].setcolor 	(gcol, pal->black);

    g[0].drawaxes 	();

    if (itype==1) g[0].marklevel(0, -1);	//mark the x-axis
    
    g[0].curve 		(*(r1),*(wfr),-1);


    //**** theta ****
    len = sprintf(str, "AT %s=%0.3g, %s=%g", "R", rexpand, "PHI", pexpand);

    g[1].setdrawarea 	(draw_area);
    g[1].setcorner 	(0, ypix-fonth-5);
    g[1].setdomain 	(0.0, 180.0);
    g[1].setscreensize 	(xpix, ypix);
    g[1].setrange 	(vmin,vmax);
    g[1].setaxislen 	(xpix/2-6*(fonth+1)-7, (ypix/2 -12 -4*fonth));

    g[1].setaxistitles 	("THETA", title(itype));
    g[1].setplottitle 	(str);

    g[1].setticlen 	(4,4);
    g[1].setcolor 	(gcol, pal->black);

    g[1].drawaxes 	();

    if (itype==1) g[1].marklevel(0, -1);       //mark the x-axis
    g[1].curve 		(*(theta1),*(wft),-1);


    //**** phi ****
    len = sprintf(str, "AT %s=%0.3g, %s=%g", "R", rexpand, "THETA", texpand);

    g[2].setdrawarea 	(draw_area);
    g[2].setcorner 	(xpix/2, ypix-fonth-5);
    g[2].setdomain 	(0.0, 360.0);
    g[2].setscreensize 	(xpix, ypix);
    g[2].setrange 	(vmin,vmax);
    g[2].setaxislen 	(xpix/2-6*(fonth+1)-7, (ypix/2 -12 -4*fonth)/2);

    g[2].setaxistitles 	("PHI", title(itype));
    g[2].setplottitle 	(str);

    g[2].setticlen 	(4, 2);
    g[2].setcolor 	(gcol, pal->black);

    g[2].drawaxes 	();

    if (itype==1) g[2].marklevel(0, -1);       //mark the x-axis
    g[2].curve 		(*(phi1),*(wfp),-1);

    SetColor(pal->white);

    if (wfscale==1000)
    {
	len=TextWidth(fonts->plottitle, "SURFACE LEVELS x1000");
	DrawText ("SURFACE LEVELS x1000", int(xpix*0.75-len/2), 35);
    }

    // write up the surface level in the main draw_area.
    WriteSurfaceLevel(wfmax0*wfscale);

    return;
}

//****************************************************************************
//****************************************************************************
void OrbitalView::Clear3d()
{
    // clear the area where the 3D representation will go
    #if DEBUGSHOW
        cout << "Clear3d\n";
    #endif
    int i;

    x0= int(xpix * 0.75);
    y0= int(ypix * 0.4);

    i= 130;
    SetColor(pal->black);
    DrawFilledBox (x0-i, y0-i, 2*i,2*i);

    return;
}

//****************************************************************************
//****************************************************************************
void OrbitalView::Setup3d()
{
    // perform necessary calculations & initialisations for the 3D view.
    #if DEBUGSHOW
	cout << "Setup3D\n";
    #endif
	
    int i, j, k;

    //**** initialize data planes to default ****
    #if DEBUG
	cout << "wf.set initialisation\n";
	cout << wf->min(1)<< "  "<<wf->min(2)<< "  "<<wf->min(3)<< "\n";
	cout << wf->max(1)<< "  "<<wf->max(2)<< "  "<<wf->max(3)<< "\n";
    #endif

    for (i=-mr; i<=mr; ++i)
    {
	for (j= -mr; j<=mr; ++j)
	{
	    for (k= -mr; k<=mr; ++k)
	    {
		wf->set(i,j,k, 1.e30);
	    }
	}
    }

    #if DEBUG
	cout << "wf.set initialisation DONE\n";
    #endif

    return;
}
//****************************************************************************
//****************************************************************************
void OrbitalView::Draw3d()
{
    #if DEBUGSHOW
	cout << "Draw3D\n";
    #endif
	

    //lots of definitions
    double
	delr, rmax,
	raxis, z, sc, delta, slev,
	a, b, c,
        // values used by the interpolation part. 
	c0, c1, c2, c3, c11, c22, c33, c12, c23, c13,
	c112, c113, c223, c122, c133, c233, c1122, c1133, c2233,
	wf00m, wf00p, wf0m0, wf0p0, wfm00, wfp00,
	wf0pp, wf0pm, wf0mp, wf0mm, wfp0p, wfp0m, wfm0p, wfm0m,
        wfpp0, wfpm0, wfmp0, wfmm0,
	ip1, ip2, jp1, jp2,
        dadx, dady, dbdx, dbdy, dcdx, dcdy, dzdx, dzdy;

    int 
        i, j, k, ip, jp,
        pcol, xp, yp, zp, ia, maxspan;

    int     alldone,nosd;

    intarray1 *onepal;
    double colourfactor;

    //******************************** 3D plot *******************************

    //**** values to evaluate xyz grid ****
    rmax= 3.0 * rexpand;
    delr= rmax / mr;
    rcut= rmax*rmax * 1.8;

    #if DEBUG
        cout << "r.set\n";
    #endif

    for (i= -mr; i<=mr; ++i)
    {
        r->set(i, i * delr);
    }

    //**** initialization, 3D axes and transformation ****
    maxspan= 0;
    sc= mcell*mr / rmax;

    #if DEBUG
        cout << "rmax= " << rmax << "\nwfmax= " << wfmax <<"\n";
        cout << "t = \n" << t;
    #endif
    //ncol= 7;
    //if (itype == 2) ncol= 13;
    //ncol= 16;

    //************************** draw 3D surface *********************

    //Unset all pixel entries:
    for (i=-(mr*mcell); i<=(mr*mcell); ++i)
    {
	for (j=-(mr*mcell); j<=(mr*mcell); ++j)
	{
	    pixset->set(i,j, -1);
	}
    }

    SetDrawMode(GXcopy);

    //**** draw lab xyz axes ****
    raxis= rmax * 1.3;

    SetColor (pal->white);

    //***draw a vector - which pixels are set is controlled by drvec . . .
    drvec (   0.0,   0.0, raxis,  "Z", x0, y0, t, sc, pixset, pal->white);
    drvec (   0.0, raxis,   0.0,  "Y", x0, y0, t, sc, pixset, pal->white);
    drvec ( raxis,   0.0,   0.0,  "X", x0, y0, t, sc, pixset, pal->white);
    drvec (   0.0,   0.0,-raxis, "\0", x0, y0, t, sc, pixset, pal->white);
    drvec (   0.0,-raxis,   0.0, "\0", x0, y0, t, sc, pixset, pal->white);
    drvec (-raxis,   0.0,   0.0, "\0", x0, y0, t, sc, pixset, pal->white);

    //**** loop over each grid cell ****
    //**** a grid cell is centred on all internal grid points only ****

    for (i= -mr+1; i<= mr-1; ++i)
    {
	for (j= -mr+1; j<=mr-1; ++j)
	{

	    //**** process down z axis till the surface is cut ****
	    for (k= mr-1; k>=-mr+1; --k)
	    {

		#if DEBUG
		    cout << "Doing rough check of surface level\n";
		#endif
		//**** rough check that its near surface ****
		c0=    wfm ( i,   j,   k   );
		wf00m= wfm ( i,   j,   k-1 );
		wf00p= wfm ( i,   j,   k+1 );
		wf0m0= wfm ( i,   j-1, k   );
		wf0p0= wfm ( i,   j+1, k   );
		wfm00= wfm ( i-1, j,   k   );
		wfp00= wfm ( i+1, j,   k   );

		a= max (c0,wf00m,wf00p,wf0m0,wf0p0,wfm00,wfp00);
		b= min (c0,wf00m,wf00p,wf0m0,wf0p0,wfm00,wfp00);

		#if DEBUG
		    cout << a << "\t" << b << "\t" << wfmax0 <<"\n";
		#endif

		if (a > 0.0)
		{
		    if ((a < 0.8*wfmax0) || (b > 1.2*wfmax0))
		    {
			#if DEBUG
			    cout << "**********max too small or min too big\n";
			    cout << (a < 0.8*wfmax0)<< "\t" << (b > 1.2*wfmax0);
			#endif
			continue;
			//cycle
		    }
		}
		else
		{
		    if (-b < 0.8*wfmax0 || -a > 1.2*wfmax0)
		    {
			#if DEBUG
			    cout << "**********max too small or min too big\n";
			    cout << (a< 0.8*wfmax0) << "\t" << (b > 1.2*wfmax0);
			#endif
			continue;
    			//cycle
		    }
		}

		#if DEBUGINT
		    cout << "Interpolating";
		#endif

		wf0pp= wfm ( i,   j+1, k+1 );
		wf0pm= wfm ( i,   j+1, k-1 );
		wf0mp= wfm ( i,   j-1, k+1 );
		wf0mm= wfm ( i,   j-1, k-1 );
		wfp0p= wfm ( i+1, j,   k+1 );
		wfm0p= wfm ( i-1, j,   k+1 );
		wfp0m= wfm ( i+1, j,   k-1 );
		wfm0m= wfm ( i-1, j,   k-1 );
		wfpp0= wfm ( i+1, j+1, k   );
		wfmp0= wfm ( i-1, j+1, k   );
		wfpm0= wfm ( i+1, j-1, k   );
		wfmm0= wfm ( i-1, j-1, k   );

		//**** interpolate wf in this cell ****
		a= 0.5 / mcell;
		c1=   (wfp00 - wfm00) * a;
		c2=   (wf0p0 - wf0m0) * a;
		c3=   (wf00p - wf00m) * a;
		a= 1.0 / (mcell*mcell);
		c11= ( (wfp00 + wfm00) * 0.5 - c0) * a;
		c22= ( (wf0p0 + wf0m0) * 0.5 - c0) * a;
		c33= ( (wf00p + wf00m) * 0.5 - c0) * a;
		a= a * 0.25;
		c12= (wfpp0 - wfpm0 - wfmp0 + wfmm0) * a;
		c13= (wfp0p - wfp0m - wfm0p + wfm0m) * a;
		c23= (wf0pp - wf0pm - wf0mp + wf0mm) * a;
		a= a / mcell;
		b= 4.0 * mcell;
		c112= (wfpp0 - wfpm0 + wfmp0 - wfmm0 - b*c2) * a;
		c113= (wfp0p - wfp0m + wfm0p - wfm0m - b*c3) * a;
		c223= (wf0pp - wf0pm + wf0mp - wf0mm - b*c3) * a;
		c122= (wfpp0 + wfpm0 - wfmp0 - wfmm0 - b*c1) * a;
		c133= (wfp0p + wfp0m - wfm0p - wfm0m - b*c1) * a;
		c233= (wf0pp + wf0pm - wf0mp - wf0mm - b*c2) * a;
		a= a / mcell;
		b= b * mcell;
		c1122= (wfpp0 + wfpm0 + wfmp0 + wfmm0 -4*c0 -b*(c11+c22)) *a;
		c1133= (wfp0p + wfp0m + wfm0p + wfm0m -4*c0 -b*(c11+c33)) *a;
		c2233= (wf0pp + wf0pm + wf0mp + wf0mm -4*c0 -b*(c22+c33)) *a;

		//**** loop over each pixel in cell ****
		alldone= TRUE;

		for (ip= -mcell2; ip <= mcell2; ++ip)
		{
		    for (jp= -mcell2; jp <= mcell2; ++jp)
		    {
			//**** check pixel isnt already set ****
			xp= i*mcell + ip;
			yp= j*mcell + jp;
			if ((pixset->get(xp,yp)!=-1) && 
				(pixset->get(xp,yp)!=pal->white) ) continue;
			#if DEBUGINT
			    cout <<"Pixel not set . . .";
			#endif

			//**** solve for wf= +/- wfmax for some z on this pixel*
			ip1= double(ip);
			ip2= ip1*ip1;
			jp1= double(jp);
			jp2= jp1*jp1;
			a= c0 + c1*ip1 + c2*jp1 + c11*ip2 + c12*ip1*jp1 +
	       		   c22*jp2 + c112*ip2*jp1 + c122*ip1*jp2 +
	       		   c1122*ip2*jp2;
			slev= wfmax0;
			if ( (a < 0.0) &&
			    ((id.n!=3) || (id.l!=0) || (wfmax0 < 0.0145)) )
        		{
        		    slev= - wfmax0;
        		}
			a -= slev;
			b= c3 + c13*ip1 + c23*jp1 + c113*ip2 + c223*jp2;
			c= c33 + c133*ip1 + c233*jp1 + c1133*ip2 + c2233*jp2;
			delta= b*b - 4.0*a*c;

			//**** find point of intersect of wf with reqd value*
			z= -100.0;
			nosd= (abs(c) < 1.e8);
			if (nosd)
        		{
			    if (b != 0.0) z= - a / b;
        		}
			else
        		{
        		    if (delta >= 0.0)
        		    {
				delta= sqrt (delta);
                		z= (-b + delta) / (2.0 * c);
                		if (z > mcell2 + 1)
                		{
                		    delta= - delta;
                		    z= (-b + delta) / (2.0 * c);
                		}
        		    }
        		}

			//**** ensure that point of int is within cell ****
			if (abs(z) <= (mcell2+1))
        		{
	      		    //**** calc derivs of z wrt x and y for 2d surface ****
			    dadx= c1 + c12*jp1 + c122*jp2 +
	      			2.0*ip1 * (c11 + c112*jp1 + c1122*jp2);
			    dady= c2 + c12*ip1 + c112*ip2 +
	      			2.0*jp1 * (c22 + c122*ip1 + c1122*ip2);
			    dbdx= c13 + c113*2.0*ip1;
			    dbdy= c23 + c223*2.0*jp1;
        		    if (nosd)
			    {
                		dzdx= - (dadx*b - a*dbdx) / (b*b);
                		dzdy= - (dady*b - a*dbdy) / (b*b);
			    }
        		    else
			    {
                		dcdx= c133 + 2.0*c1133*ip1;
                		dcdy= c233 + 2.0*c2233*jp1;
                		dzdx=((-dbdx+0.5/delta*(2*b*dbdx-4*a*dcdx-4*c*dadx))
                	    	    * c - (-b+delta)*dcdx) * 0.5 / (c*c);
                		dzdy=((-dbdy+0.5/delta*(2*b*dbdy-4*a*dcdy-4*c*dady))
                	    	    * c - (-b+delta)*dcdy) * 0.5 / (c*c);
        		    }

        		    //**** set colour as proj of 2d surface onto z ****
			    colourfactor = (1.0 + (dzdx*dzdx) + (dzdy*dzdy));
			    if (itype==1)
			    {
                                if (slev < 0.0)
                                {
                                    onepal = pal->gpal;
				    pcol   = INT (pal->ngcol / colourfactor); 
                                }
                                else
                                {
                                    onepal = pal->rpal;
				    pcol   = INT (pal->nrcol / colourfactor);
                                }
			    }
			    else
			    {
				onepal = pal->rpal;
				pcol   = INT (pal->nrcol / colourfactor);
			    }

        		    maxspan= max (maxspan,abs(i),abs(j),abs(k));

        		    //**** determine direction of greatest lobe ****

			    #if DEBUGAXESCHECK
				cout << "init " << pixset->get(xp, yp);
			    #endif

        		    //**** check if axis goes below or above ****
        		    if (pixset->get(xp, yp) == pal->white)
			    {
        			//**** determine identity of axis ****
				double zpreal = z+k*mcell;
				zp=INT(zpreal);
				vector vscre((double)xp, (double)yp, zpreal);
				vector vreal = vscre * t;

				#if DEBUGAXESCHECK
				    cout <<"\n"<< t  <<"\n"
					<< vscre  <<"\n"<< vreal <<"\n";
				#endif

				vector v1 = vreal.cross(vector(1,0,0));
				vector v2 = vreal.cross(vector(0,1,0));
				vector v3 = vreal.cross(vector(0,0,1));

				v1 = t*v1;
				v2 = t*v2;
				v3 = t*v3;

				if (int(v1.z)==0)
				{
				    //then the xaxis is in line.
				    ia=1;
				}
				else
				{
				    if (int(v2.z)==0)
			 	    {
					//then the yaxis is in line.
					ia=2;
				    }
				    else
				    {
					//then the zaxis _should_ be in line
					#if AXESCHECKSAFETY
				            if ((int(v3.z))==0)
					    {
					    	ia =3;
					    }
					    else
					    {
						#if DEBUGAXISCHECK
					            cout << "No match for axis!\n"
					            <<v1<<"\n"<<v2<<"\n"<< v3<<"\n";
	    					#endif
						if (abs(v1.z)<abs(v2.z))
						{
						    //then its v1 or v3
						    if (abs(v1.z)<abs(v3.z))
						    {
							ia=1;
						    }
						    else
						    {
							ia=3;
						    }
						}
						else
						{
						    //else its v2 or v3
						    if (abs(v2.z)<abs(v3.z))
						    {
							ia=2;
						    }
						    else
						    {
							ia=3;
						    }
						}
					    }
					#else
					    ia=3;
					#endif
				    }
				}

				#if DEBUGAXISCHECK
				    cout << v1 <<"\n"<< v2<<"\n" << v3<<"\n";
				    cout << "\tia = "<<ia<<"\n";
				#endif

                                z= (t.elem(1,ia)*xp + t.elem(2,ia)*yp) *
                                     t.elem(3,ia)/(1.0 - pow(t.elem(3,ia),2)); 

                                if (z<zpreal)
				{
				    SetColor (onepal->get(pcol));
                                    pixset->set(xp,yp, onepal->get(pcol));
                            	    #if DEBUGPALUSE
                                	cout <<"\tfinal " << pcol << "\n";
                                    #endif
                                    DrawPixel (x0+xp,y0-yp);
				}
			    }
			    else
			    {
				SetColor (onepal->get(pcol));
                                pixset->set(xp,yp, onepal->get(pcol));
                                #if DEBUGPALUSE
                                    cout <<"\tfinal "  << pcol << "\n";
                                #endif
                                DrawPixel (x0+xp,y0-yp);

        		    }//if pcol
			}
			else
			{
	  		    //**** no hit for this pixel ****
			    alldone= FALSE;
			}
		    }
		}//if abs(z) <= mcell2 +1

		//**** only do next lower z cell if not all pixels set ****
		if (alldone) break;
		//exit
	    }//k
	    SyncDisplay();
	}
    }

    //******* finished 3D drawing ******

    SetColor(pal->black);
    return;
}

// *************************************************************************
// *************************************************************************
void OrbitalView::Redraw3d()
{
    // a _FAST_ redraw of the screen 
    // axis labels will not be redrawn - but that's OK

    int i, j;
    for (i=-(mr*mcell); i<=(mr*mcell); ++i)
    {
	for (j=-(mr*mcell); j<=(mr*mcell); ++j)
	{
	    SetColor(pixset->get(i,j));
	    DrawPixel(x0+i, y0-j);
	}
    }
}

// *************************************************************************
// *************************************************************************
void OrbitalView::DrawLev()
{
    //draw the level of the 3D plots and setup the mousedown CB

    #if DEBUGSHOW
	cout << "DrawLev\n";
    #endif
    lastwfval=wfmax0;
 
    //**** mark surface level ****
    dlev3d (wfmax0*wfscale, pal->rpal->get(20), pal->gpal->get(20));

    SetButtonDownCB(draw_area, wfcontrolMouseDownCB);
}

//******************************************************************************
//******************************************************************************
double
OrbitalView::hwf (double x, double y, double z)
{
    // calculates the H wavefunction of the current orbital

    #if DEBUGWF
	cout << "In hwf\n";
    #endif

   //**** evaluates H-atom wavefunction type qta (n,l,m) at posn (x,y,z) ****

    double	r,alfa,qz,qa,h0,costh,sinth,xx,
	    	cosphi,sinphi,cosmphi,sinmphi;

    double	hwfi;	    // hold the val of hwf before it is returned

    int		i;

    r= sqrt (x*x + y*y + z*z);
    qz= 1.0;
    alfa= 1.0;
    qa= qz * alfa;
    h0= radwf(id.n,id.l,r,qa) / rtpi;

    if ((id.l == 0) || (r == 0.0))
    {
	hwfi= 0.0;
	if (id.l == 0) hwfi= h0 * 0.5;
    }
    else
    {
	costh= z / r;
	if (costh >= 1.0) costh=  1.0;
	if (costh <= -1.0) costh= -1.0;
	if ( (id.m != 0) && ( abs(costh) == 1.0))
	{
	    hwfi= 0.0;
	}
	else
	{
	    sinth= sqrt (1.0- (costh*costh) );
	    h0= h0 * angwf (id.l,abs(id.m),costh,sinth);
	    //**** multiply by PHI wavefunction ****
	    if (id.m == 0)
	    {
		hwfi= h0 / rt2;
	    }
	    else
	    {
		cosphi= x / (r * sinth);
		sinphi= y / (r * sinth);
		cosmphi= cosphi;
		sinmphi= sinphi;
		for (i= 2; i<=abs(id.m); ++i)
		{
		    xx=      cosmphi*cosphi - sinmphi*sinphi;
		    sinmphi= cosmphi*sinphi + sinmphi*cosphi;
		    cosmphi= xx;
		}
		if (id.m > 0) cosmphi= sinmphi;
		hwfi= h0 * cosmphi;
	    }
	}
    }

    return hwfi;
}

// *************************************************************************
// *************************************************************************
double OrbitalView::angwf (int l, int m, double costh, double sinth)
{
    //**** norm. angular wf THETA using Associated Legendre Polynomials ****

    #if DEBUGWF
	cout << "in angwf (l,m) ="<< l << " " <<m<<"  \n";
	cout << "costh "<< costh <<"  sinth "<<sinth<<"\n";
    #endif
    double	legen;


    double	angwfi;		    //temp to hold returning value

    int	    i,j;

    if ((m != 0) && (abs(sinth) <= 1.e-5))
    {
	return (0.0);
    }

    // **** Legendre poly, do all l for m= 0 ****
    #if DEBUGWF
	cout << "calc legen\n";
    #endif
    p->set(0,0, 1.0);
    p->set(1,0, costh);
    for (i=1; i<=(l+m-1); ++i)
    {
	//cout << "iterating on p.set()\t";
	double temp = ( (2*i+1)*costh*p->get(i,0) - i*p->get(i-1,0) ) / (i+1);
	//cout << temp << "\n";
	p->set(i+1,0, temp) ;
    }

    //cout << p <<"\n";

    if (m == 0)
    {
        legen= p->get(l,0);
    }
    else
    {
	// **** Associated Legendre poly, increase m ****
	//cout <<"Second iteration start . . .\n";
	//cout << "(l,m) = "<< l << " " <<m<<"  \n";
	for ( j= 1; j<=m; ++j)
	{
	    //cout << "(l,m) = ("<< l << ", " <<m<<")\n";
	    //cout << "j = "<< j << "\n";
	    //cout << (l-m+j) << "  " << (l+m-j);
            for (i= (l-m+j); i<=(l+m-j); ++i)
	    {
	        //cout << "i = "<< i << "\n";
		p->set(i,j, (( (i+j-1)*(i+j  ) * p->get(i-1,j-1) -
        		      (i-j+1)*(i-j+2) * p->get(i+1,j-1) ) / ((2*i+1) * sinth) ));
	    }
	}
	legen= p->get(l,m);
	//cout <<"Second iteration over . . .\n";
    }
    //cout << p <<"\n";

    //**** mormalize ****
    angwfi= legen * sqrt ( (2*l+1)/2.0*fact(l-m)/fact(l+m) );

    #if DEBUGWF
	//cout << "legen = " <<legen << "\n";
	cout << "angwfi = " <<angwfi << "\n";
    #endif

    return angwfi;
}

// *************************************************************************
// *************************************************************************
double OrbitalView::radwf(int n, int l, double r, double alfa)
{
    //**** this is Laguerre polynomial L(n+l,2*l+1,rho) times H wf norm *****

    double      lag, rho, signx, radwfi;
    int	        k;

    rho= 2.0 * r * alfa / n;

    // **** evaluate Laguerre polynomial ****
    signx= -1.0;
    lag= 0.0;
    for ( k= 0; k<= n-l-1; ++k)
    {
	lag= lag + signx * fact(n+l)*fact(n+l) /
		(fact(n-l-1-k) * fact(2*l+1+k) * fact(k) );
	signx= - signx * rho;
    }

    // **** normalize ****
    radwfi= - lag * sqrt( pow( (2.0*alfa/(n *fact(n+l))), 3)  / (2 * n)
    		* fact(n-l-1) ) * exp(-rho/2.0) * (pow(rho, l));

    #if DEBUGWF
	cout << "radwfi = " <<radwfi << "\n";
    #endif
    return radwfi;
}
//*****************************************************************************
//*****************************************************************************
double OrbitalView::wfm (int i0, int j, int k)
{
    //**** evaluates/retrieves wavefunction at given 3d grid point ****

    #if DEBUGWF
	cout << "In wfm\n";
    #endif

    double x,y,z, x1,y1,z1, wf0,a;

    if (wf->get(i0,j,k) == 1.0e30)		//1e30 is the dummy value
    {
	#if DEBUGWF
	    cout <<"Need to calc\n";
	#endif
	x= r->get(i0);
	y= r->get(j);
	z= r->get(k);
	#if DEBUGWF
	    printf("(x,y,z) = (%f, %f, %f)\n", x, y, z);
	    cout << t;
	#endif
	//*** back transform to get world x,y,z ****
	x1= t.getcol(1).dot(vector(x, y, z));
	y1= t.getcol(2).dot(vector(x, y, z));
	z1= t.getcol(3).dot(vector(x, y, z));
	wf0= 0.0;
	a= x1*x1 + y1*y1 + z1*z1;
	#if DEBUGWF
	    cout <<"a = "<<a<<"\n";
	#endif
	if (a <= rcut) wf0= hwf (x1,y1,z1);
	if (itype == 2) wf0= wf0*wf0;
	wf->set(i0,j,k, wf0);
    }
    #if DEBUGWF
	cout << "wf = "<<wf->get(i0,j,k) << "\n";
    #endif
    return wf->get(i0,j,k);
}

//*****************************************************************************
//*****************************************************************************
int OrbitalView::WhichGraph (int x, int y)
{
    // works out which graph the mouse is in for the WFval selection
    // returns TRUE if in a graph, otherwise FALSE
    int i=2;
    while ( (i>=0) &&  
              (     (g[i].getblx() > x) || (g[i].gettrx() < x) ||  
                    (g[i].getbly() < y) || (g[i].gettry() > y)      )  
          ) i--;  

    currentgraph = i;
    return(TRUE&&(currentgraph+1));
}

//*****************************************************************************
//*****************************************************************************
void OrbitalView::DrawLevPix (int y)
{
    //draws the level of the graph based on a y-pixel val
    #if DEBUGDYNALEVEL
        cout << "lastwf = " << lastwfval;
    #endif

    double newwfval =  (g[currentgraph].getfnval(y))/wfscale;
    #if DEBUGDYNALEVEL
	cout << " currentgraph: " << currentgraph 
		<< " newwf = " << newwfval;
    #endif

    newwfval = MAX( newwfval,g[currentgraph].getymin() );

    dlev3d (lastwfval*wfscale, pal->rpal->get(20), pal->gpal->get(20));

    lastwfval = abs(newwfval);
    if (lastwfval==0) lastwfval = g[currentgraph].getymax()*0.0001/wfscale;

    #if DEBUGDYNALEVEL
	cout << " newwf = " << lastwfval << "\n";
    #endif

    dlev3d (lastwfval*wfscale, pal->rpal->get(20), pal->gpal->get(20));

    WriteSurfaceLevel(lastwfval*wfscale);

    return;
}
//*****************************************************************************
//*****************************************************************************
void OrbitalView::WriteSurfaceLevel(double d)
{
    // (remove old and) write the wf val inthe draw_area

    SetWidgetFont(draw_area, fonts->plottitle);
    SetColor(pal->white);
    DrawText ("                             ", int(xpix*0.75-50), 50);
    int len = sprintf(str, "SURFACE LEVEL=%f", d);
    len=TextWidth(fonts->plottitle, str);
    DrawText (str, int(xpix*0.75-len/2), 50);
    return;
}

//******************************************************************************
//******************************************************************************
void OrbitalView::dlev3d (double wfmax, int colour1, int colour2)
{
    //**** this draws 3d shell value on 1D plots ****

    int iplot;

    SetDrawMode(SANE_XOR);

    #if DEBUGDLEV
        printf ("Dlev3d: wfmax=%g, col=%d\n", wfmax, colour);
    #endif

    for (iplot= 0; iplot <= 2; ++iplot)
    {
        //**** +ve level ****
        g[iplot].marklevel (wfmax, colour1);

        //**** -ve level ****
        if (itype==1)
        {
            g[iplot].marklevel (-wfmax, colour2);
        }
    }
    #if DEBUGDLEV
        printf ("Dlev3d: out\n");
    #endif

    SetDrawMode(GXcopy);

    return;
}

//*****************************************************************************
//*****************************************************************************
void drvec (double x, double y, double z, char* lab, int x0, int y0,
            matrix t, double scale, intarray2 *pixset, int colour)
{
    //**** draws a vector with label on 3D plot ****
    //given the vector, the label to put at the end, the transformation
    //matrix, the pixel/world scale, the colour, and any pixel setting
    //information to be kept track of.
    //
    // if (pixset==NULL) then pixel setting information is not saved

    double x1,y1,z1;
    int	i;

    //**** forward transform of vector ****
    vector v (x, y, z);
    x1= t.getrow(1).dot(v);
    y1= t.getrow(2).dot(v);
    z1= t.getrow(3).dot(v);


    //**** draw axis and label ****
    DrawText(lab, (x0+int(x1*scale*1.1)), (y0-int(y1*scale*1.1)));

    if (pixset)
    {
	//*** put the line info into *(pixset) ***
	if (int(x1)!=0)
	{

	    double m=y1/x1;
	    if (abs(m)<1)
	    {
		for (i=0; i<=(int(x1*scale)); ++i)
		{
	    	    pixset->set(i, INT(i*m), colour);
	    	    DrawPixel (x0+i, y0- INT(i*m));
		}
        	for (i=0; i>=(int(x1*scale)); --i)
        	{
            	    pixset->set(i, INT(i*m), colour);
            	    DrawPixel (x0+i, y0- INT(i*m));
        	}
	    }
	    else
	    {
		for (i=0; i<=(int(y1*scale)); ++i)
        	{
                    pixset->set(INT(i/m),i, colour);
                    DrawPixel (x0+INT(i/m), y0-i); 
        	}
        	for (i=0; i>=(int(y1*scale)); --i)
        	{
                    pixset->set(INT(i/m),i, colour);
                    DrawPixel (x0+INT(i/m), y0-i);
        	}
	    }
	}
	else
	{
	    for (i=0; i<=(int(y1*scale)); ++i)
	    {
		pixset->set(0, i, colour);
		DrawPixel (x0, y0-i);
	    }
            for (i=0; i>=(int(y1*scale)); --i)
            {
        	pixset->set(0, i, colour);
        	DrawPixel (x0, y0-i);
            }

	}
    }
    else
    {
	DrawLine (x0, y0, (x0+int(x1*scale)), (y0-int(y1*scale)) ); 
    }

    return;
}


//*****************************************************************************
//*****************************************************************************
double axmax (double wfmax)
{
    //**** calcuates a suitable y scale to fit data with given maximum ****

    double dmax;
    int	i;
    
    i= int (log10 (wfmax*1.e10)) - 10;

    dmax= wfmax / pow(10.0, i);

    //dmax = ceil( log10(dmax)/log10(2) );
    if (dmax > 8)
    {
        dmax= 16.0;
    }
    else
    {
        if (dmax > 4)
        {
            dmax= 8.0;
        }
        else
        {
            if (dmax > 2)
            {
                dmax= 4.0;
            }
            else
            {
                dmax= 2.0;
            }
        }
    }

    return (dmax * pow(10, i));
}

//******************************************************************************
//******************************************************************************
void TitleRedisplay (Widget w, int width, int height, void *data)
{
    // redisplay CB function for the title area
    // NOTE: may also be caled to trip an update of information

    SetBgColor(w, pal->black);
    ClearDrawArea();
    SetColor (pal->white);

    OrbitalView *me = (OrbitalView *)data;

    if ( *(me->choosemode) != WAITING_FOR_SELECTION )
    {

	int x=150/2;
	int y=25;

    	SetWidgetFont(w, fonts->title);
	
	if (me->itype==1)
	{
	    char *s;
	    s="WAVE";
	    DrawText  (s, x-TextWidth(fonts->title, s)/2, y);
	    s="FUNCTION";
	    y+=FontHeight(fonts->title);
	    DrawText  (s, x-TextWidth(fonts->title, s)/2, y);
	}
	else
	{
	    sprintf (me->str, "%s", title(me->itype));
	    DrawText  (me->str, x-TextWidth(fonts->title, me->str)/2, y);
	}

	y+=40;

	//me->id.str
	x-=int(TextWidth(fonts->Stitle, me->id.str)/2);
	sprintf (me->str, "%d", me->id.n);
	DrawText(me->str, x, y);
	DrawTextH (w,TRUE, me->id.str, x+TextWidth(fonts->Stitle, me->str) ,y);
	y+=40;
	x=75;

    	SetWidgetFont(w, fonts->title);

	sprintf (me->str, "n = %d", me->id.n);
	DrawText(me->str, x-TextWidth(fonts->title,me->str)/2, y);
	y+=25;
	sprintf (me->str, "l = %d", me->id.l);
	DrawText(me->str, x-TextWidth(fonts->title,me->str)/2, y);
	y+=25;
	if (me->id.m == 0)
	{
	    sprintf (me->str, "|m| = 0");
	}
	else
	{
	    if (me->id.m > 0)
            {
        	sprintf (me->str, "|m| = %d (%s)", me->id.m, "sin");
            }
            else
            {
        	sprintf (me->str, "|m| = %d (%s)",-me->id.m, "cos");
	    }
	}
	DrawText(me->str, x-TextWidth(fonts->title,me->str)/2, y);
	
    }
    else
    {
	SetWidgetFont(w, fonts->title);
	int x=150/2;
	int y=60;
	char *s;
	s="PLEASE";
	DrawText  (s, x-TextWidth(fonts->title, s)/2, y);
	y+=30;
	s="SELECT";
	DrawText  (s, x-TextWidth(fonts->title, s)/2, y);
	y+=30;
	s="AN";
	DrawText  (s, x-TextWidth(fonts->title, s)/2, y);
	y+=30;
	s="ORBITAL";
	DrawText  (s, x-TextWidth(fonts->title, s)/2, y);
    }
}
